/**
 * @file	cf_thread.h
 * @author	myusgun <myusgun@gmail.com>
 * @version	0.1
 *
 * @remark 멀티스레드 및 뮤텍스 지원
 */
#ifndef __CF_THREAD_H__
#define __CF_THREAD_H__

#include "cf_base.h"

#define CF_ERROR_THREAD_CREATE			CF_ERROR_THREAD - 1
#define CF_ERROR_THREAD_INVALID_ARGS	CF_ERROR_THREAD - 2

#define CF_ERROR_MUTEX_CREATE			CF_ERROR_MUTEX - 1
#define CF_ERROR_MUTEX_INVALID_ARGS		CF_ERROR_MUTEX - 2

#ifdef _WIN32
# define CF_THREAD_RETURN	unsigned int	/**< 스레드 워커 함수 반환 형 */
# define CF_THREAD_CALL		__stdcall		/**< 스레드 워커 함수 호출 규약 */
#else // #ifdef _WIN32
# define CF_THREAD_RETURN	void *			/**< 스레드 워커 함수 반환 형 */
# define CF_THREAD_CALL						/**< 스레드 워커 함수 호출 규약 */
#endif // #ifdef _WIN32

/** 스레드 워커 함수 형태이며, CF_Thread_Function 형식 */
typedef CF_THREAD_RETURN  (CF_THREAD_CALL * CF_Thread_Function) (void *);

/** 스레드 아이디 */
typedef void *	CF_Thread;

/** 뮤텍스 아이디 */
typedef void *	CF_Mutex;

#ifdef __cplusplus
extern "C" {
#endif

/**
 * 스레드를 생성
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param threadID	스레드 아이디 주소
 * @param callback	스레드 워커 함수 이름
 * @param arg		스레드 함수로 전달할 인자
 *
 * @code

#include "cf_thread.h"

CF_Mutex globalMutex;

CF_THREAD_RETURN CF_THREAD_CALL worker (void * arg) {
	if (CF_Mutex_Create (&globalMutex) < 0) {
		// error
	}

	if (CF_Mutex_Lock (&globalMutex) < 0) { // for critical section
		// error
	}

	// critical section

	if (CF_Mutex_Unlock (&globalMutex) < 0) { // for critical section
		// error
	}

	if (CF_Mutex_Destory (&globalMutex) < 0) {
		// error
	}
}

int main (void) {
	CF_Thread tid;
	if (CF_Thread_Create (&tid, worker, NULL) < 0) {
		// error
	}

	if (CF_Thread_Join (&tid) < 0) { // block
		// error
	}

	if (CF_Thread_Release (&tid) < 0) {
		// error
	}
}

 * @endcode
 */
CF_EXPORT int
CF_Thread_Create	(CF_Thread			* threadID,
					 CF_Thread_Function	callback,
					 void				* arg);

/**
 * 스레드 아이디를 해제
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param threadID 스레드 아이디 주소
 *
 * @remark 스레드 아이디를 해제하는 것이며 워커 스레드가 종료되지 않음
 */
CF_EXPORT int
CF_Thread_Release	(CF_Thread * threadID);

/**
 * 스레드가 종료될 때 까지 대기
 *
 * @return CF_OK 반환
 *
 * @param threadID 스레드 아이디 주소
 */
CF_EXPORT int
CF_Thread_Join		(CF_Thread * threadID);

/**
 * 뮤텍스 생성
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param mutex 뮤텍스 아이디 주소
 *
 * @see CF_Thread_Create
 */
CF_EXPORT int
CF_Mutex_Create		(CF_Mutex * mutex);

/**
 * 뮤텍스 해제
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param mutex 뮤텍스 아이디 주소
 */
CF_EXPORT int
CF_Mutex_Destory	(CF_Mutex * mutex);

/**
 * 뮤텍스 잠금
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param mutex 뮤텍스 아이디 주소
 */
CF_EXPORT int
CF_Mutex_Lock		(CF_Mutex * mutex);

/**
 * 뮤텍스 잠금 해제
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param mutex 뮤텍스 아이디 주소
 */
CF_EXPORT int
CF_Mutex_Unlock		(CF_Mutex * mutex);

#ifdef __cplusplus
}
#endif

#endif // #ifndef __CF_THREAD_H__
